/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2010 UCM
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author:  Delfín Rupérez Cañas <delfinrc@fdi.ucm.es>
 *          
 */

#include "ns3/assert.h"

#include "antor-dlr-header.h"
#include "ns3/log.h"

#define IPV4_ADDRESS_SIZE 4
#define ANTOR_MSG_HEADER_SIZE 8
#define ANTOR_DELAY_SIZE 4

namespace ns3 {

namespace antordlr {

NS_LOG_COMPONENT_DEFINE("AntORDNRHeader");


// ---------------- ANTOR Message -------------------------------

NS_OBJECT_ENSURE_REGISTERED ( MessageHeader );

MessageHeader::MessageHeader() :
	m_messageType(MessageHeader::MessageType(0)) {
}

MessageHeader::~MessageHeader() {
}

TypeId MessageHeader::GetTypeId(void) {
	static TypeId tid =
			TypeId("ns3::antordlr::MessageHeader")
                              .SetParent<Header> () 
                              .AddConstructor<MessageHeader> ();
	return tid;
}
TypeId MessageHeader::GetInstanceTypeId(void) const {
	return GetTypeId();
}

uint32_t MessageHeader::GetSerializedSize(void) const {
	uint32_t size = ANTOR_MSG_HEADER_SIZE;
	switch (m_messageType) {
	case HELLO_MESSAGE:
		NS_LOG_DEBUG("Hello Message Size: " << size << " + "
				<< m_message.hello.GetSerializedSize());
		size += m_message.hello.GetSerializedSize();
		break;
	case RFA_MESSAGE:
		NS_LOG_DEBUG("Reactive Forward Ant Message Size: " << size << " + "
				<< m_message.rfa.GetSerializedSize());
		size += m_message.rfa.GetSerializedSize();
		break;
	case RBA_MESSAGE:
		NS_LOG_DEBUG("Reactive Backward Ant Message Size: " << size << " + "
				<< m_message.rba.GetSerializedSize());
		size += m_message.rba.GetSerializedSize();
		break;
	case PFA_MESSAGE:
		NS_LOG_DEBUG("Proactive Forward Ant Message Size: " << size << " + "
				<< m_message.pfa.GetSerializedSize());
		size += m_message.pfa.GetSerializedSize();
		break;
	case PBA_MESSAGE:
		NS_LOG_DEBUG("Proactive Backward Ant Message Size: " << size << " + "
				<< m_message.pba.GetSerializedSize());
		size += m_message.pba.GetSerializedSize();
		break;
	case LFN_MESSAGE:
		NS_LOG_DEBUG("Link Failure Notification Message Size: " << size
				<< " + " << m_message.lfn.GetSerializedSize());
		size += m_message.lfn.GetSerializedSize();
		break;
	case RRFA_MESSAGE:
		NS_LOG_DEBUG("Route Repair Forward Ant Message Size: " << size << " + "
				<< m_message.rrfa.GetSerializedSize());
		size += m_message.rrfa.GetSerializedSize();
		break;
	case RRBA_MESSAGE:
		NS_LOG_DEBUG("Route Repair Backward Ant Message Size: " << size
				<< " + " << m_message.rrba.GetSerializedSize());
		size += m_message.rrba.GetSerializedSize();
		break;
	case ULN_MESSAGE:
		NS_LOG_DEBUG("Unicast link notification Message Size: " << size
						<< " + " << m_message.uln.GetSerializedSize());
				size += m_message.uln.GetSerializedSize();
				break;
	default:
		NS_ASSERT(false);
	}
	return size;
}

void MessageHeader::Print(std::ostream &os) const {
	// TODO
}

void MessageHeader::Serialize(Buffer::Iterator start) const {
	Buffer::Iterator i = start;
	i.WriteU8(m_messageType);
	i.WriteU8(m_numBroadcast);
	i.WriteHtonU16(GetSerializedSize());
	i.WriteU8(m_timeToLive);
	i.WriteU8(m_hopCount);
	i.WriteHtonU16(m_messageSequenceNumber);

	switch (m_messageType) {
	case HELLO_MESSAGE:
		m_message.hello.Serialize(i);
		break;
	case RFA_MESSAGE:
		m_message.rfa.Serialize(i);
		break;
	case RBA_MESSAGE:
		m_message.rba.Serialize(i);
		break;
	case PFA_MESSAGE:
		m_message.pfa.Serialize(i);
		break;
	case PBA_MESSAGE:
		m_message.pba.Serialize(i);
		break;
	case LFN_MESSAGE:
		m_message.lfn.Serialize(i);
		break;
	case RRFA_MESSAGE:
		m_message.rrfa.Serialize(i);
		break;
	case RRBA_MESSAGE:
		m_message.rrba.Serialize(i);
		break;
	case ULN_MESSAGE:
		m_message.uln.Serialize(i);
		break;

	default:
		NS_ASSERT(false);
	}

}

uint32_t MessageHeader::Deserialize(Buffer::Iterator start) {
	uint32_t size;
	Buffer::Iterator i = start;
	m_messageType = (MessageType) i.ReadU8();
	m_numBroadcast = i.ReadU8();
	m_messageSize = i.ReadNtohU16();
	m_timeToLive = i.ReadU8();
	m_hopCount = i.ReadU8();
	m_messageSequenceNumber = i.ReadNtohU16();

	size = ANTOR_MSG_HEADER_SIZE;
	switch (m_messageType) {
	case HELLO_MESSAGE:
		size += m_message.hello.Deserialize(i, m_messageSize
				- ANTOR_MSG_HEADER_SIZE);
		break;
	case RFA_MESSAGE:
		size += m_message.rfa.Deserialize(i, m_messageSize
				- ANTOR_MSG_HEADER_SIZE);
		break;
	case RBA_MESSAGE:
		size += m_message.rba.Deserialize(i, m_messageSize
				- ANTOR_MSG_HEADER_SIZE);
		break;
	case PFA_MESSAGE:
		size += m_message.pfa.Deserialize(i, m_messageSize
				- ANTOR_MSG_HEADER_SIZE);
		break;
	case PBA_MESSAGE:
		size += m_message.pba.Deserialize(i, m_messageSize
				- ANTOR_MSG_HEADER_SIZE);
		break;
	case LFN_MESSAGE:
		size += m_message.lfn.Deserialize(i, m_messageSize
				- ANTOR_MSG_HEADER_SIZE);
		break;
	case RRFA_MESSAGE:
		size += m_message.rrfa.Deserialize(i, m_messageSize
				- ANTOR_MSG_HEADER_SIZE);
		break;
	case RRBA_MESSAGE:
		size += m_message.rrba.Deserialize(i, m_messageSize
				- ANTOR_MSG_HEADER_SIZE);
		break;
	case ULN_MESSAGE:
		size += m_message.uln.Deserialize(i, m_messageSize
				- ANTOR_MSG_HEADER_SIZE);
		break;



	default:
		NS_ASSERT(false);
	}
	return size;
}

// ---------------- ANTOR HELLO Message -------------------------------

uint32_t MessageHeader::Hello::GetSerializedSize(void) const {
	return (2 * this->destValue.size() * IPV4_ADDRESS_SIZE)
			+ ANTOR_DELAY_SIZE;

}

void MessageHeader::Hello::Print(std::ostream &os) const {
	// TODO
}

void MessageHeader::Hello::Serialize(Buffer::Iterator start) const {
	ns3::Buffer::Iterator i = start;

	uint8_t *buf = (uint8_t *) &this-> delay;
	for (uint32_t j = 0; j < sizeof(float); ++j, ++buf) {
		i.WriteU8(*buf);
	}

	//  i.WriteU32(this->delay);
	for (size_t n = 0; n < this->destValue.size(); ++n) {
		i.WriteHtonU32(this->destValue[n].address.Get());

		uint8_t *buf2 = (uint8_t *) &this->destValue[n].best;
		for (uint32_t j2 = 0; j2 < sizeof(float); ++j2, ++buf2) {
			i.WriteU8(*buf2);
		}

	}

}

uint32_t MessageHeader::Hello::Deserialize(Buffer::Iterator start,
		uint32_t messageSize) {
	Buffer::Iterator i = start;

	NS_ASSERT((messageSize - ANTOR_DELAY_SIZE) % (IPV4_ADDRESS_SIZE * 2)
			== 0);
	int numAddresses = (messageSize - ANTOR_DELAY_SIZE) / IPV4_ADDRESS_SIZE
			/ 2;

	float v;
	uint8_t *buf = (uint8_t *) &v;
	for (uint32_t k = 0; k < sizeof(float); ++k, ++buf) {
		*buf = i.ReadU8();
	}

	this -> delay = v;
	this->destValue.clear();
	for (int n = 0; n < numAddresses; ++n) {
		Ipv4Address address(i.ReadNtohU32());
		float v2;
		uint8_t *buf2 = (uint8_t *) &v2;
		for (uint32_t k2 = 0; k2 < sizeof(float); ++k2, ++buf2) {
			*buf2 = i.ReadU8();
		}
		float best = v2;
		this->destValue.push_back((DestinationValue) {
			address, best
		});
	}
	return messageSize;
}

// ---------------- ANTOR RFA Message -------------------------------

uint32_t MessageHeader::Rfa::GetSerializedSize(void) const {
	return (this->listP.size() * IPV4_ADDRESS_SIZE) + (IPV4_ADDRESS_SIZE * 2);

}

void MessageHeader::Rfa::Print(std::ostream &os) const {
	// TODO
}

void MessageHeader::Rfa::Serialize(Buffer::Iterator start) const {
	ns3::Buffer::Iterator i = start;
	i.WriteHtonU32(this->addrSession.source.Get());
	i.WriteHtonU32(this->addrSession.destination.Get());
	for (size_t n = 0; n < this->listP.size(); ++n)
		i.WriteHtonU32(this->listP[n].Get());
}

uint32_t MessageHeader::Rfa::Deserialize(Buffer::Iterator start,
		uint32_t messageSize) {
	Buffer::Iterator i = start;

	NS_ASSERT((messageSize - (IPV4_ADDRESS_SIZE * 2)) % (IPV4_ADDRESS_SIZE)
			== 0);
	int numAddresses = (messageSize - (IPV4_ADDRESS_SIZE * 2))
			/ IPV4_ADDRESS_SIZE;
	Ipv4Address address1(i.ReadNtohU32());
	Ipv4Address address2(i.ReadNtohU32());
	this -> addrSession.source = address1;
	this -> addrSession.destination = address2;
	this->listP.clear();
	for (int n = 0; n < numAddresses; ++n) {
		Ipv4Address address(i.ReadNtohU32());
		this->listP.push_back(address);
	}
	return messageSize;
}

// ---------------- ANTOR RBA Message -------------------------------

uint32_t MessageHeader::Rba::GetSerializedSize(void) const {
	return (this->listP.size() * IPV4_ADDRESS_SIZE) + (IPV4_ADDRESS_SIZE * 2)
			+ ANTOR_DELAY_SIZE;

}

void MessageHeader::Rba::Print(std::ostream &os) const {
	// TODO
}

void MessageHeader::Rba::Serialize(Buffer::Iterator start) const {
	ns3::Buffer::Iterator i = start;

	// Writing delay
	uint8_t *buf = (uint8_t *) &this-> delay;
	for (uint32_t j = 0; j < sizeof(float); ++j, ++buf) {
		i.WriteU8(*buf);
	}

	i.WriteHtonU32(this->addrSession.source.Get());
	i.WriteHtonU32(this->addrSession.destination.Get());

	for (size_t n = 0; n < this->listP.size(); ++n)
		i.WriteHtonU32(this->listP[n].Get());
}

uint32_t MessageHeader::Rba::Deserialize(Buffer::Iterator start,
		uint32_t messageSize) {
	Buffer::Iterator i = start;

	NS_ASSERT((messageSize - (IPV4_ADDRESS_SIZE * 2) - ANTOR_DELAY_SIZE)
			% (IPV4_ADDRESS_SIZE) == 0);
	int numAddresses = (messageSize - (IPV4_ADDRESS_SIZE * 2)
			- ANTOR_DELAY_SIZE) / IPV4_ADDRESS_SIZE;

	// Reading delay
	float v;
	uint8_t *buf = (uint8_t *) &v;
	for (uint32_t k = 0; k < sizeof(float); ++k, ++buf) {
		*buf = i.ReadU8();
	}

	this -> delay = v;

	Ipv4Address address1(i.ReadNtohU32());
	Ipv4Address address2(i.ReadNtohU32());
	this -> addrSession.source = address1;
	this -> addrSession.destination = address2;
	this->listP.clear();
	for (int n = 0; n < numAddresses; ++n) {
		Ipv4Address address(i.ReadNtohU32());
		this->listP.push_back(address);
	}
	return messageSize;
}

// ---------------- ANTOR PFA Message -------------------------------

uint32_t MessageHeader::Pfa::GetSerializedSize(void) const {
	return (this->listP.size() * IPV4_ADDRESS_SIZE) + (IPV4_ADDRESS_SIZE * 2);

}

void MessageHeader::Pfa::Print(std::ostream &os) const {
	// TODO
}

void MessageHeader::Pfa::Serialize(Buffer::Iterator start) const {
	ns3::Buffer::Iterator i = start;
	i.WriteHtonU32(this->addrSession.source.Get());
	i.WriteHtonU32(this->addrSession.destination.Get());
	for (size_t n = 0; n < this->listP.size(); ++n)
		i.WriteHtonU32(this->listP[n].Get());
}

uint32_t MessageHeader::Pfa::Deserialize(Buffer::Iterator start,
		uint32_t messageSize) {
	Buffer::Iterator i = start;

	NS_ASSERT((messageSize - (IPV4_ADDRESS_SIZE * 2)) % (IPV4_ADDRESS_SIZE)
			== 0);
	int numAddresses = (messageSize - (IPV4_ADDRESS_SIZE * 2))
			/ IPV4_ADDRESS_SIZE;
	Ipv4Address address1(i.ReadNtohU32());
	Ipv4Address address2(i.ReadNtohU32());
	this -> addrSession.source = address1;
	this -> addrSession.destination = address2;
	this->listP.clear();
	for (int n = 0; n < numAddresses; ++n) {
		Ipv4Address address(i.ReadNtohU32());
		this->listP.push_back(address);
	}
	return messageSize;
}

// ---------------- ANTOR PBA Message -------------------------------

uint32_t MessageHeader::Pba::GetSerializedSize(void) const {
	return (this->listP.size() * IPV4_ADDRESS_SIZE) + (IPV4_ADDRESS_SIZE * 2)
			+ ANTOR_DELAY_SIZE;

}

void MessageHeader::Pba::Print(std::ostream &os) const {
	// TODO
}

void MessageHeader::Pba::Serialize(Buffer::Iterator start) const {
	ns3::Buffer::Iterator i = start;

	// Writing delay
	uint8_t *buf = (uint8_t *) &this-> delay;
	for (uint32_t j = 0; j < sizeof(float); ++j, ++buf) {
		i.WriteU8(*buf);
	}

	i.WriteHtonU32(this->addrSession.source.Get());
	i.WriteHtonU32(this->addrSession.destination.Get());

	for (size_t n = 0; n < this->listP.size(); ++n)
		i.WriteHtonU32(this->listP[n].Get());
}

uint32_t MessageHeader::Pba::Deserialize(Buffer::Iterator start,
		uint32_t messageSize) {
	Buffer::Iterator i = start;

	NS_ASSERT((messageSize - (IPV4_ADDRESS_SIZE * 2) - ANTOR_DELAY_SIZE)
			% (IPV4_ADDRESS_SIZE) == 0);
	int numAddresses = (messageSize - (IPV4_ADDRESS_SIZE * 2)
			- ANTOR_DELAY_SIZE) / IPV4_ADDRESS_SIZE;

	// Reading delay
	float v;
	uint8_t *buf = (uint8_t *) &v;
	for (uint32_t k = 0; k < sizeof(float); ++k, ++buf) {
		*buf = i.ReadU8();
	}

	this -> delay = v;

	Ipv4Address address1(i.ReadNtohU32());
	Ipv4Address address2(i.ReadNtohU32());
	this -> addrSession.source = address1;
	this -> addrSession.destination = address2;
	this->listP.clear();
	for (int n = 0; n < numAddresses; ++n) {
		Ipv4Address address(i.ReadNtohU32());
		this->listP.push_back(address);
	}
	return messageSize;
}

// ---------------- ANTOR LFN Message -------------------------------

uint32_t MessageHeader::Lfn::GetSerializedSize(void) const {
	//return (this->listP.size() * IPV4_ADDRESS_SIZE) + (this->dest.size()
	//		* IPV4_ADDRESS_SIZE) + (2*IPV4_ADDRESS_SIZE) ;
    return (this->dest.size()*IPV4_ADDRESS_SIZE) + (2*IPV4_ADDRESS_SIZE);
}

void MessageHeader::Lfn::Print(std::ostream &os) const {
	// TODO
}

void MessageHeader::Lfn::Serialize(Buffer::Iterator start) const {
	ns3::Buffer::Iterator i = start;
	i.WriteHtonU16(this->reserved);
	i.WriteHtonU16(this->ldest);
    i.WriteHtonU32(this->originator.Get());

	for (size_t n = 0; n < this->ldest; ++n) {
		i.WriteHtonU32(this->dest[n].Get());

	}
}

uint32_t MessageHeader::Lfn::Deserialize(Buffer::Iterator start,
		uint32_t messageSize) {
	Buffer::Iterator i = start;

	NS_ASSERT( (messageSize - (2*IPV4_ADDRESS_SIZE)) % IPV4_ADDRESS_SIZE == 0);
	this->reserved = i.ReadNtohU16();
	this -> ldest = i.ReadNtohU16();
	Ipv4Address originatorAddr(i.ReadNtohU32());
	this->originator = originatorAddr;
	int numAddresses = (messageSize - (2*IPV4_ADDRESS_SIZE)) / IPV4_ADDRESS_SIZE;

	this->dest.clear();
	for (int n = 0; n < numAddresses; ++n) {
		Ipv4Address address(i.ReadNtohU32());
		this->dest.push_back(address);
	}
	return messageSize;
}


// ---------------- ANTOR RRFA Message -------------------------------

uint32_t MessageHeader::Rrfa::GetSerializedSize(void) const {
	return (this->listP.size() * IPV4_ADDRESS_SIZE) + (IPV4_ADDRESS_SIZE * 3);

}

void MessageHeader::Rrfa::Print(std::ostream &os) const {
	// TODO
}

void MessageHeader::Rrfa::Serialize(Buffer::Iterator start) const {
	ns3::Buffer::Iterator i = start;
	i.WriteHtonU32(this->addrSession.source.Get());
	i.WriteHtonU32(this->addrSession.destination.Get());
	i.WriteHtonU32(this->addrSession.local.Get());
	for (size_t n = 0; n < this->listP.size(); ++n)
		i.WriteHtonU32(this->listP[n].Get());
}

uint32_t MessageHeader::Rrfa::Deserialize(Buffer::Iterator start,
		uint32_t messageSize) {
	Buffer::Iterator i = start;

	NS_ASSERT((messageSize - (IPV4_ADDRESS_SIZE * 3)) % (IPV4_ADDRESS_SIZE)
			== 0);
	int numAddresses = (messageSize - (IPV4_ADDRESS_SIZE * 3))
			/ IPV4_ADDRESS_SIZE;
	Ipv4Address address1(i.ReadNtohU32());
	Ipv4Address address2(i.ReadNtohU32());
	Ipv4Address address3(i.ReadNtohU32());
	this -> addrSession.source = address1;
	this -> addrSession.destination = address2;
	this -> addrSession.local = address3;
	this->listP.clear();
	for (int n = 0; n < numAddresses; ++n) {
		Ipv4Address address(i.ReadNtohU32());
		this->listP.push_back(address);
	}
	return messageSize;
}

// ---------------- ANTOR RRBA Message -------------------------------

uint32_t MessageHeader::Rrba::GetSerializedSize(void) const {
	return (this->listP.size() * IPV4_ADDRESS_SIZE) + (IPV4_ADDRESS_SIZE * 2)
			+ ANTOR_DELAY_SIZE;

}

void MessageHeader::Rrba::Print(std::ostream &os) const {
	// TODO
}

void MessageHeader::Rrba::Serialize(Buffer::Iterator start) const {
	ns3::Buffer::Iterator i = start;

	// Writing delay
	uint8_t *buf = (uint8_t *) &this-> delay;
	for (uint32_t j = 0; j < sizeof(float); ++j, ++buf) {
		i.WriteU8(*buf);
	}

	i.WriteHtonU32(this->addrSession.source.Get());
	i.WriteHtonU32(this->addrSession.destination.Get());


	for (size_t n = 0; n < this->listP.size(); ++n)
		i.WriteHtonU32(this->listP[n].Get());
}

uint32_t MessageHeader::Rrba::Deserialize(Buffer::Iterator start,
		uint32_t messageSize) {
	Buffer::Iterator i = start;

	NS_ASSERT((messageSize - (IPV4_ADDRESS_SIZE * 2) - ANTOR_DELAY_SIZE)
			% (IPV4_ADDRESS_SIZE) == 0);
	int numAddresses = (messageSize - (IPV4_ADDRESS_SIZE * 2)
			- ANTOR_DELAY_SIZE) / IPV4_ADDRESS_SIZE;

	// Reading delay
	float v;
	uint8_t *buf = (uint8_t *) &v;
	for (uint32_t k = 0; k < sizeof(float); ++k, ++buf) {
		*buf = i.ReadU8();
	}

	this -> delay = v;

	Ipv4Address address1(i.ReadNtohU32());
	Ipv4Address address2(i.ReadNtohU32());


	this -> addrSession.source = address1;
	this -> addrSession.destination = address2;
	this->listP.clear();
	for (int n = 0; n < numAddresses; ++n) {
		Ipv4Address address(i.ReadNtohU32());
		this->listP.push_back(address);
	}
	return messageSize;
}


// ---------------- ANTOR ULN Message -------------------------------

uint32_t MessageHeader::Uln::GetSerializedSize(void) const {
	return 2*IPV4_ADDRESS_SIZE;

}

void MessageHeader::Uln::Print(std::ostream &os) const {
	// TODO
}

void MessageHeader::Uln::Serialize(Buffer::Iterator start) const {
	ns3::Buffer::Iterator i = start;


	i.WriteHtonU32(this->dst.Get());
	i.WriteHtonU32(this->src.Get());
}

uint32_t MessageHeader::Uln::Deserialize(Buffer::Iterator start,
		uint32_t messageSize) {
	Buffer::Iterator i = start;

	Ipv4Address address1(i.ReadNtohU32());
	Ipv4Address address2(i.ReadNtohU32());
	this ->dst = address1;
	this ->src = address2;
	return messageSize;
}



} // namespace antor-dlr
} // namespace ns3


