/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2010 UCM
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author:  Delfín Rupérez Cañas <delfinrc@fdi.ucm.es>
 *          
 */

#ifndef ANTOR_DLR_ROUTING_TABLE_H_
#define ANTOR_DLR_ROUTING_TABLE_H_


#include <stdint.h>
#include <cassert>
#include <map>
#include <sys/types.h>
#include "ns3/ipv4.h"
#include "ns3/ipv4-route.h"
#include "ns3/timer.h"
#include "ns3/net-device.h"
#include "ns3/random-variable.h"
#include "ns3/output-stream-wrapper.h"
#include <algorithm>
#include <iomanip>
#include "ns3/simulator.h"
#include "ns3/log.h"

namespace ns3 {

namespace antordlr {

typedef struct Phero_Table_Val {

    double pheroReg;
    double pheroVir;
    double numAvgHop;


} PheroTable;

/**
 * \ingroup antHocNet
 * \brief Routing table entry
 */
class RoutingTableEntry
{
public:

  /// c-to
	 /// c-to
  RoutingTableEntry (Ptr<NetDevice> dev = 0,Ipv4Address dst = Ipv4Address (), Ipv4InterfaceAddress iface = Ipv4InterfaceAddress (), uint16_t  hops = 0,
	                 Ipv4Address nextHop = Ipv4Address (), PheroTable f_t = PheroTable(), bool b = false);

  ~RoutingTableEntry ();
  
 
  Ipv4Address GetDestination () const { return m_ipv4Route->GetDestination(); }
  Ptr<Ipv4Route> GetRoute () const { return  m_ipv4Route; }
  void SetRoute (Ptr<Ipv4Route> r) { m_ipv4Route = r; }
  void SetNextHop (Ipv4Address nextHop) { m_ipv4Route->SetGateway(nextHop); }
  Ipv4Address GetNextHop () const { return m_ipv4Route->GetGateway(); }
  void SetOutputDevice (Ptr<NetDevice> dev) { m_ipv4Route->SetOutputDevice(dev); }
  Ptr<NetDevice> GetOutputDevice () const { return m_ipv4Route->GetOutputDevice(); }
  Ipv4InterfaceAddress GetInterface () const { return m_iface;}
  void SetInterface (Ipv4InterfaceAddress iface) { m_iface = iface; }

  void SetHop (uint16_t hop) { m_hops = hop; }
  uint16_t GetHop () const {return m_hops; }
  void SetPheroTable(PheroTable pt) { m_pheroTable = pt;} 
  PheroTable GetPheroTable() { return m_pheroTable; } 

  bool GetVisitedLink() { return m_visitedlink; }
  void SetVisitedLink(bool b) { m_visitedlink = b;}
  ///\name Precursors management
    //\{
    /**
     * Insert precursor in precursor list if it doesn't yet exist in the list
     * \param id precursor address
     * \return true on success
     */
    bool InsertPrecursor (Ipv4Address id);
    /**
     * Lookup precursor by address
     * \param id precursor address
     * \return true on success
     */
    bool LookupPrecursor (Ipv4Address id);
    /**
     * \brief Delete precursor
     * \param id precursor address
     * \return true on success
     */
    bool DeletePrecursor (Ipv4Address id);
    /// Delete all precursors
    void DeleteAllPrecursors ();
    /**
     * Check that precursor list empty
     * \return true if precursor list empty
     */
    bool IsPrecursorListEmpty () const;
    /**
     * Inserts precursors in vector prec if they does not yet exist in vector
     */
    void GetPrecursors (std::vector<Ipv4Address> & prec) const;
    //\}

  /**
   * \brief Compare destination address
   * \return true if equal
   */
  bool operator== (Ipv4Address const  dst) const
  {
    return (m_ipv4Route->GetDestination () == dst);
  }

private:
 
  /// Hop Count (number of hops needed to reach destination)
  uint16_t m_hops;

  /** Ip route, include
  *   - destination address
  *   - source address
  *   - next hop address (gateway)
  *   - output device
  */
  Ptr<Ipv4Route> m_ipv4Route;
  /// Output interface address
  Ipv4InterfaceAddress m_iface;

  /// List of precursors
   std::vector<Ipv4Address> m_precursorList;

  // Pheromone Table
  PheroTable m_pheroTable;

  bool m_visitedlink; ///< Visited links.
 };

/**
 * \ingroup anthocnet
 * \brief The Routing table used by ANTOR-DNR protocol
 */
class RoutingTable
{
public:

  RoutingTable () {}
  

   //\}
  // Table Routing Manage
 
  /// Delete all entries from routing table
  void Clear () { m_table.clear (); }

  // Get routing table size
  uint32_t GetSize() {return m_table.size(); }
  
  /// Delete all route from interface with address iface
   void DeleteAllRoutesFromInterface (Ipv4InterfaceAddress iface);

  /**
    * Add routing table entry if it doesn't yet exist in routing table
    * \param r routing table entry
    * \return true in success
    */
    bool AddRoute (RoutingTableEntry & r);

  /**
    * Delete routing table entry with destination address and index idx.
    * \param idx index in routing table
    * \return true on success
    */
    bool DeleteRoute (Ipv4Address dest, uint32_t idx);


    /**
      * Lookup routing table entries with destination address dst
      * \param dst destination address
      * \param rts entries with destination address dst, if exists
      * \return true on success
      */
    bool LookupAllRoutes (Ipv4Address dst, std::vector<RoutingTableEntry>& rts);

    /**
      * Lookup VALID routing table entries with destination address dst
      * \param dst destination address
      * \param rts entries with destination address dst, if exists
      * \return true on success
      */
    bool LookupRoutesWithReg (Ipv4Address dst, std::vector<RoutingTableEntry>&  rts);

    /// Update routing table
    bool Update (RoutingTableEntry & rt, uint32_t idx);

    /// Update routing table
    bool UpdateNew (RoutingTableEntry & rt);


    RoutingTableEntry GetRoutingEntry(uint32_t idx);

    bool GetIndexRoute (Ipv4Address dst, Ipv4Address next, uint32_t& idx);

    bool LookupEntriesByDest (Ipv4Address dest,std::vector<uint32_t> &idxs) const;

    bool LookupEntryByDestAndNextHop (Ipv4Address dest,Ipv4Address next, uint32_t &idx) const;

    bool LookupEntriesByNextHop (Ipv4Address next, std::vector<uint32_t> &idx) const;

    // Destination Selection Manage
    std::vector<Ipv4Address> GetAllDestination(uint32_t max);

    bool FindDestinationAux(Ipv4Address dest, std::vector<Ipv4Address> temp);

 
 // Destination of Link Failure Notification
 std::vector<Ipv4Address> GetDestinationLFN(Ipv4Address dir);

 std::vector<Ipv4Address> GetSources(Ipv4Address dst, Ipv4Address nexthop);

 // Lookup minimum Hop in Proactive Process
 bool LookupMinHop(Ipv4Address dest,RoutingTableEntry&);

 void
   Purge ();

private:
  std::multimap<Ipv4Address,RoutingTableEntry> m_table;
  

};



} // namespace antordlr

} // namespace ns3

#endif /* ANTOR_DNR_ROUTING_TABLE_H_ */



