/*
 * Copyright 2012 IDSIA
 *
 *
 * Author:  Michal Kudelski
 *
 * (based on packet-sink.h and .cc)
 *
 */

#ifndef MAPPLE_APP_H
#define MAPPLE_APP_H

#include "ns3/application.h"
#include "ns3/event-id.h"
#include "ns3/ptr.h"
#include "ns3/traced-callback.h"
#include "ns3/address.h"
#include "ns3/ipv4-interface-container.h"
#include "ns3/node-container.h"
#include "ns3/net-device.h"
#include "ns3/random-variable.h"
#include "ns3/mobility-model.h"
#include "ns3/waypoint-mobility-model.h"
#include "ns3/waypoint.h"
#include "ns3/wifi-mac-queue.h"
#include <iostream>
#include <string>
//#include <stdlib.h>
#include <stdio.h>

#include <cmath>
#include "svm.h"

namespace ns3 {

//////////////////////////////////////////////
// Forward declarations
//////////////////////////////////////////////
class Address;
class Socket;
class Packet;

//////////////////////////////////////////////
// Defines
//////////////////////////////////////////////

// Parameters //////////////////////////////////////////
#define MAX_REPORTED_NEIGHBORS 50
#define MAX_NODE_ID 100

#define SEND_UNICAST 0

#define ANNOUNCE_TIMES 3
#define RANDOM_STATE_CHANGE 1
#define PROBING_PACKET_SIZE 1000
#define NO_MOVE 0

#define SAVE_PROCESSED_SAMPLES 1
//#define SAVE_SAMPLES_GEO 1
//#define SAVE_WAYPOINTS 1
#define VALIDATION 1

#define  P_MOVE 0.6	//0.1
#define  P_DISABLED 0.2	//0.25
#define  P_PROBE 0.2	//0.65

// NORMAL TRAFFIC (for gathering samples)
#define  RATE_LOW_MU 0.1 // 100 ms
#define  RATE_LOW_SIGMA 0.005

#define  RATE_MED_MU 0.025 // 25 ms
#define  RATE_MED_SIGMA 0.005

#define  RATE_HIGH_MU 0.0033 // 3.3 ms
#define  RATE_HIGH_SIGMA 0.005

//// LOW TRAFFIC (for testing purposes)
//#define  RATE_LOW_MU 1.0 // 1000 ms
//#define  RATE_LOW_SIGMA 0.25
//
//#define  RATE_MED_MU 0.75 // 750 ms
//#define  RATE_MED_SIGMA 0.25
//
//#define  RATE_HIGH_MU 0.5 // 500 ms
//#define  RATE_HIGH_SIGMA 0.1

#define  MAX_NEIGHBOR_TTL 2.0  // seconds
#define  CLEANING_TIMER_PERIOD 1.0  // seconds
#define  MIN_SAMPLE_RECEIVED 20		// by default, 50

// mobility parameters
#define ASSUMED_COMMUNICATION_RANGE 100.0
#define P_GO_MORE	0.5
#define MIN_DISTANCE 200.0	// 100.0
#define MAX_TRIALS 8
#define MAX_DISTANCE 500.0	// 300.0
#define STEP_FACTOR 50.0
#define AREA_MAX_X 600
#define AREA_MAX_Y 600
#define SPEED 25.0	// in [m/s], 10.0 m/s by default, 5 m/s in the new training procedure

// sample processing parameters
#define NUMBER_OF_DIST_INTERVALS	2		// number of distance intervals
#define INTERVAL_LIMIT_1			70.0	// first distance interval limit

// precomputed normalization values (SVM PREDICTION)
static const double featuremaxM1[] = {121.25,-47.016,3,3,9,4,5,4,5,5,5,10,4,5,4,5,5,5,46.586};
static const double featureminM1[] = {4.0922,-90.632,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,3.2137};
static const double featuremaxM2[] = {121.25,-47.016,3,3,9,4,5,4,5,5,5,10,4,5,4,5,5,5};
static const double featureminM2[] = {4.0922,-90.632,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
static const double featuremaxM3[] = {121.25,3,3,9,4,5,4,5,5,5,10,4,5,4,5,5,5};
static const double featureminM3[] = {4.0922,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
static const double featuremaxM4[] = {-47.016,3,3,9,4,5,4,5,5,5,10,4,5,4,5,5,5};
static const double featureminM4[] = {-90.632,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
static const double featuremaxM5[] = {3,3,9,4,5,4,5,5,5,10,4,5,4,5,5,5};
static const double featureminM5[] = {1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

static const double featuremaxMnew[] = {125.892738000000,-37.0688290000000,3,3,10,5,6,5,5,5,6,11,5,6,5,5,5,6,56.5957180000000};
static const double featureminMnew[] = {1.90717700000000,-90.8461670000000,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,3.00871800000000};

// Structures //////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
// estimators' data structures
////////////////////////////////////////////////////////////////////////////////////

// moving window average
typedef struct moving_window_struct {
	std::vector<double> window;
	unsigned int window_size;
	unsigned int current_size;
	double current_sum;

	moving_window_struct() : window_size(1), current_size(0), current_sum(0.0) {}

	void update(double value) {
		if (current_size == window_size) {
			current_sum = current_sum - window[0] + value;
			window.erase(window.begin());
			window.push_back(value);
		} else {
			current_size++;
			current_sum += value;
			window.push_back(value);
		}
	}
} moving_window_t;

// exponential weighted moving average (ewma)
typedef struct ewma_struct {
	double alpha;
	double current_value;

	ewma_struct() : alpha(0.0), current_value(0.0) {}

	void update(double value) {
		current_value = alpha * current_value + (1 - alpha) * value;
	}
} ewma_t;

typedef struct ewma_prr_struct {
	double alpha;
	double current_value;

	uint32_t last_sequence_number;
	double timestamp;

	double current_rate;

	ewma_prr_struct() : alpha(0.0), current_value(0.0), last_sequence_number(0), timestamp(0.0), current_rate(0.0) {}

	void update(uint32_t number_of_missed, uint32_t number_of_received, uint32_t sequence_number) {
		for (unsigned int i = 0; i < number_of_missed; i++ )
			current_value = alpha * current_value;
		for (unsigned int i = 0; i < number_of_received; i++ )
			current_value = alpha * current_value + (1 - alpha);
		if (number_of_received > 0) {
			last_sequence_number = sequence_number;
			// timestamp = Simulator::Now().GetSeconds();
		}
	}
} ewma_prr_t;

typedef struct wmemwa_prr_struct {
	double time_interval;
	double alpha;
	double current_value;

	uint32_t last_sequence_number;
	double timestamp;

	uint32_t received;
	uint32_t failed;

	wmemwa_prr_struct() : time_interval(0.0), alpha(0.0), current_value(0.0), last_sequence_number(0), timestamp(0.0), received(0), failed(0) {}

	void update(uint32_t number_of_missed, uint32_t number_of_received, uint32_t sequence_number) {
		failed += number_of_missed;
		received += number_of_received;
		if (number_of_received > 0) {
			last_sequence_number = sequence_number;
			// timestamp = Simulator::Now().GetSeconds();
		}
	}

	void periodic_update() {
		if ( (received == 0) && (failed == 0) )
			return;
		current_value = alpha * current_value + (1 - alpha) * ( 1.0 * received / (received + failed) );
		received = 0;
		failed = 0;
	}
} wmemwa_prr_t;
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////

/*size: 144 bytes...*/
typedef struct probe_msg {
	uint8_t msg_type;
	uint32_t counter;
	uint8_t source;
	uint32_t n_samples;
	uint32_t last_update;
	uint8_t state;
	uint8_t rate;
	uint32_t source_status[5];
	float rssi_data[MAX_REPORTED_NEIGHBORS];
	uint8_t rate_data[MAX_REPORTED_NEIGHBORS];
	double sendTime;
} probe_msg_t;

typedef struct subsample {
	// timestamp
	double timestamp;

	// measured values
	double rrsi;
	double sinr;
	double distance;

	// estimated values
	//double MAPPLE_estimator;
	double singleRSSI_estimator;
	double singleSNIR_estimator;
	double mwRSSI_estimator;
	double mwSNIR_estimator;
	double ewmaRSSI_estimator;
	double ewmaSNIR_estimator;
	double ewmaPRR_estimator;
	double wmewmaPRR_estimator;

} subsample_t;

typedef struct sample
{
	uint32_t sender_status[5];
	uint32_t receiver_status[5];
	uint8_t sender_rate;
	uint32_t received;
	uint32_t first_counter;
	uint32_t last_counter;
	float sender_rssi_data[MAX_REPORTED_NEIGHBORS];
	uint8_t sender_rate_data[MAX_REPORTED_NEIGHBORS];
	double last_sample_time;
	double avgTravelTime;
	double avgSendTime;
	double avgSendTimeWithoutBackoff;
	uint32_t queueSize;
	std::vector<subsample_t> subsamples;
} sample_t;

typedef struct neighbor_tentry
{
	double distance;
	double sum_rssi;
	double sum_snr;
	uint32_t rssi_samples;
	uint8_t state;
	uint8_t rate;
	uint32_t last_update;
	double last_message_time;

	// estimators here
	double last_rssi;
	double last_snir;
	moving_window_t mw_rssi;
	moving_window_t mw_snir;
	ewma_t ewma_rssi;
	ewma_t ewma_snir;
	ewma_prr_t ewma_prr;
	wmemwa_prr_t wmemwa_prr;
} neighbor_entry_t;

// Constants //////////////////////////////////////////
#define  AM_UNIT_MSG 0x89
#define  STATE_DISABLED 1
#define  STATE_PROBING 2
#define  STATE_SAMPLING 2
#define  STATE_MOVING 3
#define  STATE_NULL 4
#define  MSG_PROBE 1
#define  MSG_ANNOUNCE 2

#define  RATE_NULL 0
#define  RATE_LOW 1
#define  RATE_MED 2
#define  RATE_HIGH 3

#define MAPPLE_PI_VALUE 3.14159265358979323846


// Other defines //////////////////////////////////////////
#define is_set(X,Y) (X & (1 << (Y)))
#define set_bit(X,Y) X = ((X) | (1 << (Y)))
#define clear_bit(X,Y) X = ((X) & ~(1 << (Y)))

////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////

/**
 * \ingroup applications
 * \defgroup packetsink PacketSink
 *
 * This application was written to implement MAPPLE: it gathers link quality estimates
 */

/**
 * \ingroup MAPPLE
 *
 * \brief This application was written to implement MAPPLE: it gathers link quality estimates
 *
 * This application was written to implement MAPPLE: it gathers link quality estimates.
 *
 */
class MappleApplication : public Application
{
public:
  static TypeId GetTypeId (void);
  MappleApplication ();

  virtual ~MappleApplication ();

  /**
   * \return the total bytes received in this sink app
   */
  uint32_t GetTotalRx () const;

  /**
   * \return pointer to listening socket
   */
  Ptr<Socket> GetListeningSocket (void) const;

  /**
   * \return list of pointers to accepted sockets
   */
  std::list<Ptr<Socket> > GetAcceptedSockets (void) const;


  void SetInterfaceContainer(Ipv4InterfaceContainer inter){this->interfaces = inter;}

  void SetNodeContainer(NodeContainer nc){this->nodes = nc;}

  void SetPrefix(std::string prefix) {this->outFilePath = prefix;}

  void SetModel(std::string model) {this->modelName = model;}

protected:
  virtual void DoDispose (void);
private:
  // inherited from Application base class.
  virtual void StartApplication (void);    // Called at time specified by Start
  virtual void StopApplication (void);     // Called at time specified by Stop

  // Events handlers for various timer-triggered events
  void SendPacket ();			// A new packet is generated and send to MAC layer
  void CleaningEvent ();		// Neighbors table is cleaned
  void ChangeStateEvent();		// Here is where nodes change states

  // Other events handlers
  void HandleRead (Ptr<Socket>);
  bool HandlePromiscuous(Ptr<NetDevice>, Ptr<const Packet>, uint16_t, const Address &, const Address &, enum NetDevice::PacketType);	// Called for packets successfully received by the device, in promiscuous mode

  void HandleAccept (Ptr<Socket>, const Address& from);
  void HandlePeerClose (Ptr<Socket>);
  void HandlePeerError (Ptr<Socket>);

  // Other helpful functions
  void ProcessPacketData(probe_msg_t*, Ipv4Address, double, double, double, double, uint32_t);	// Performs the packet data processing, after the packet is read via 'HandleRead'
  void changeState(uint8_t newState);		// Handle the state change
  void initializeSampling();				// Initialize sample table
  void startProbing();  					// Select rate and start probing
  void stopProbing();						// StopProbing
  void Move();					 			// Move the node
  void print_sample(uint8_t id);			// Output sample. Increment n_samples
  void closeSample(uint8_t id);				// close a single sample
  void closeAllSamples(); 					// Just call close for all
  void clearMacQueue();						// Clean the MAC interface queue
  uint8_t next_rate();						// Determines randomly PROBE rate
  double next_message_interval();			// This one determines time between messages depending on node's current rate
  uint8_t nextState();	// This function determines the state transitions
  double stateChangeInterval();				// Determines the time interval to call the state change
  double next_gaussian(double mu, double sigma);		// Draw a normally distributed random variable (given the mean and the std deviation of the normal distribution)
  void announceState(uint8_t announcementsNumber);		// Announce state change to neighbors
  int choose_destination();						// Chooses the destination for unicast transmissions, when in UNICAST probing mode

  Vector getNextDestination();								// MoveNodeStrategy2 implementation
  int numberNeighbors(Vector position, double range);		// how many neighbors are we going to have in the given position, assuming the given range of transmission

  void updateEstimatorsOnReceiveBeforeProcessing(probe_msg_t*, double, double);
  void updateEstimatorsOnReceiveAfterProcessing(probe_msg_t*, double, double);
  void updateEstimatorsOnReceive(probe_msg_t* msg, double d1, double d2) {
	  updateEstimatorsOnReceiveBeforeProcessing(msg, d1, d2);
	  updateEstimatorsOnReceiveAfterProcessing(msg, d1, d2);
  }
  void updateEstimatorsPeriodic();

  ////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  // DATA RELATED TO prediction using SVM model
  //
  ////////////////////////////////////////////////////////////////////////////////////////////////////
  std::string modelfilename;		// path to the above file (for writing samples)

  static std::string modelName;

  int nooffeatures;
  int svm_type;
  int nr_class;

  static uint32_t* classCounter;

  static double* featuremax;
  static double* featuremin;
  // flag indicating which features are used for validation
  static bool useDistance;
  static bool useRSSI;
  static bool useSNR;

  // model parameters
  struct svm_model* model;
  struct svm_node *testfeatures;

  static double errSum;
  static uint32_t errCounter;


  ////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  // DATA RELATED TO ns3 application code
  //
  ////////////////////////////////////////////////////////////////////////////////////////////////////

  // In the case of TCP, each socket accept returns a new socket, so the
  // listening socket is stored seperately from the accepted sockets
  Ptr<Socket>     m_socket;       // Listening socket
  std::list<Ptr<Socket> > m_socketList; //the accepted sockets
  Ptr<Socket> sendSocket;		// Socket for sending data

  uint32_t        m_totalRx;      // Total bytes received
  TypeId          m_tid;          // Protocol TypeId
  TracedCallback<Ptr<const Packet>, const Address &> m_rxTrace;		// a trace source to detect received packages
  TracedCallback<Ptr<const Packet> > m_txTrace;						// a trace source to detect transmitted packages
  Ipv4InterfaceContainer interfaces;	// known interfaces
  NodeContainer nodes;	// known interfaces
  uint16_t portBase;

  EventId         m_sendEvent;    		// Eventid of pending "send packet" event
  EventId         m_cleanEvent;   	 	// Eventid of pending "clean" event
  EventId         m_stateChangeEvent;   // Eventid of pending "state-change" event

  Ptr<WaypointMobilityModel> m_mobility;	// mobility model of this node
  //Ptr<MobilityModel> m_mobility;	// mobility model of this node

  Ptr< WifiMacQueue > macQueue;		// pointer to the MAC queue

  ////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  // DATA RELATED TO MAPPLE
  //
  ////////////////////////////////////////////////////////////////////////////////////////////////////

	// output file
	static FILE* fp;				// file for writing samples
	std::string outFilePath;		// path to the above file (for writing samples)
	static FILE* fp_samples_geo;	// file for writing geographical coordinates related to samples
	static FILE* fp_mobility;		// file for the whole mobility trace
	static FILE* fp_processed;		// file for writing processed samples

	// samples counter
	static uint32_t sampleID;

	// Neighbor table
	// Look: neighbor_entry_t struct
	// in header file
	// MAX_NODE_ID-> To use static size table
	neighbor_entry_t Neighbors[MAX_NODE_ID];

	// Flag which indicates that node is
	// probing
	// Update: Seems redundant given that:
	// probe <=> (currentState == STATE_PROBING)
	bool probe;

	// Variables used by the random number generator
	NormalVariable normalVariable;
	UniformVariable uniformVariable;

	// Table containing all collected samples
	sample_t samples[MAX_NODE_ID];


	uint32_t my_status[5];  // Status of all my neighbors as bit array (probing or not)
	uint32_t last_update;   // Last msg_counter at which an update was received
	uint32_t n_samples;     // Number of samples flushed
	uint32_t my_counter;

	// My state
	uint8_t currentState;
	uint8_t oldState;
	uint8_t currentRate;
	int currentDestination;

	// State transition timer settings
	// Transition time is uniform random (min, max)
	double min_state_change_time; // secs
	double max_state_change_time; // secs

  ////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  // end of: DATA RELATED TO MAPPLE
  //
  ////////////////////////////////////////////////////////////////////////////////////////////////////

};

} // namespace ns3

#endif /* MAPPLE_APP_H */

