#include "pfr-routing-protocol.h"

#include "ns3/random-variable.h"

#include "ns3/boolean.h"

#include "ns3/log.h"

NS_LOG_COMPONENT_DEFINE ("PfrRoutingProtocol");

namespace ns3
{
namespace pfr
{
NS_OBJECT_ENSURE_REGISTERED (RoutingProtocol);

/// UDP Port for PFR control traffic
//const uint32_t RoutingProtocol::PFR_PORT = 654;


RoutingProtocol::RoutingProtocol () :
 Parameter1(0),
 BroadcastProbability(0.5),
 EnableBroadcast(false),
 IdCacheMemoryTime(10.0),
 m_dpd (Seconds(IdCacheMemoryTime))
{
	printf("Hello Constructor PFR !!! \n");
}

TypeId
RoutingProtocol::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::pfr::RoutingProtocol")
    .SetParent<Ipv4RoutingProtocol> ()
    .AddConstructor<RoutingProtocol> ()
    .AddAttribute ("Parameter1", "Parameter1.",
               UintegerValue (10),
               MakeUintegerAccessor (&RoutingProtocol::Parameter1),
               MakeUintegerChecker<uint32_t> ())
    .AddAttribute("BroadcastProbability", "Probability that the node will broadcast a packet (when performing probabilistic flooding)",
	    DoubleValue(0.5),
	    MakeDoubleAccessor(&RoutingProtocol::BroadcastProbability),
	    MakeDoubleChecker<double>(0.0,1.0))
	.AddAttribute("IdCacheMemoryTime", "How long packet IDs are going to be remembered to avoid routing loops during flooding...",
		DoubleValue(10.0),
		MakeDoubleAccessor(&RoutingProtocol::IdCacheMemoryTime),
		MakeDoubleChecker<double>(10.0))
	.AddAttribute ("EnableBroadcast", "Indicates whether a broadcast data packets forwarding enable.",
               BooleanValue (false),
               MakeBooleanAccessor (&RoutingProtocol::SetBroadcastEnable,
                                    &RoutingProtocol::GetBroadcastEnable),
               MakeBooleanChecker ())
  ;
  return tid;
}

void
RoutingProtocol::SetParameter1 (uint32_t p) {
	Parameter1 = p;
}

void RoutingProtocol::SetBroadcastProbability (double bp) {
	BroadcastProbability = bp;
}

RoutingProtocol::~RoutingProtocol ()
{
}

void
RoutingProtocol::DoDispose ()
{
  Ipv4RoutingProtocol::DoDispose ();
}

void
RoutingProtocol::PrintRoutingTable (Ptr<OutputStreamWrapper> stream) const
{
  *stream->GetStream () << "Node: " << m_ipv4->GetObject<Node> ()->GetId () << " Time: " << Simulator::Now ().GetSeconds () << "s ";
  //m_routingTable.Print (stream);
  std::ostream* os = stream->GetStream ();
  *os << "Routing table..." << std::endl;
}

void
RoutingProtocol::SetIpv4 (Ptr<Ipv4> ipv4)
{
  NS_ASSERT (ipv4 != 0);
  NS_ASSERT (m_ipv4 == 0);

  m_ipv4 = ipv4;

  // Create lo route. It is asserted that the only one interface up for now is loopback
  NS_ASSERT (m_ipv4->GetNInterfaces () == 1 && m_ipv4->GetAddress (0, 0).GetLocal () == Ipv4Address ("127.0.0.1"));
  m_lo = m_ipv4->GetNetDevice (0);
  NS_ASSERT (m_lo != 0);

//  // Remember lo route  <<<=========----------  TODO !!!!
//  RoutingTableEntry rt (/*device=*/ m_lo, /*dst=*/ Ipv4Address::GetLoopback (), /*know seqno=*/ true, /*seqno=*/ 0,
//                                    /*iface=*/ Ipv4InterfaceAddress (Ipv4Address::GetLoopback (), Ipv4Mask ("255.0.0.0")),
//                                    /*hops=*/ 1, /*next hop=*/ Ipv4Address::GetLoopback (),
//                                    /*lifetime=*/ Simulator::GetMaximumSimulationTime ());
//  m_routingTable.AddRoute (rt);
//
//  Simulator::ScheduleNow (&RoutingProtocol::Start, this);
}

Ptr<Ipv4Route>
RoutingProtocol::RouteOutput (Ptr<Packet> p, const Ipv4Header &header,
                              Ptr<NetDevice> oif, Socket::SocketErrno &sockerr)
{
	//	  if (!p)
	//	    {
	//	      return LoopbackRoute (header, oif); // later
	//	    }
	//	  if (m_socketAddresses.empty ())
	//	    {
	//	      sockerr = Socket::ERROR_NOROUTETOHOST;
	//	      NS_LOG_LOGIC ("No aodv interfaces");
	//	      Ptr<Ipv4Route> route;
	//	      return route;
	//	    }

	sockerr = Socket::ERROR_NOTERROR;

	Ipv4InterfaceAddress ifaceAddr = m_ipv4->GetAddress (1, 0);		// assuming only one device, and only one address on it
	Ptr<NetDevice> dev = m_ipv4->GetNetDevice (1);

	// If we try to send packet to ourselves...
	if (header.GetDestination () == ifaceAddr.GetLocal ()) {
		  NS_ASSERT (m_lo != 0);
		  Ptr<Ipv4Route> rt = Create<Ipv4Route> ();
		  rt->SetDestination (header.GetDestination ());
		  rt->SetSource (ifaceAddr.GetLocal ());
		  NS_ASSERT_MSG (rt->GetSource () != Ipv4Address (), "Valid PFR source address not found");
		  rt->SetGateway (Ipv4Address ("127.0.0.1"));
		  rt->SetOutputDevice (m_lo);
		  return rt;
	}

	// Return broadcast as a first route
	Ptr<Ipv4Route> route;
	route = Create<Ipv4Route> ();
	route->SetDestination (header.GetDestination ());
	//route->SetGateway (header.GetDestination ());		// TODO temporary solution for '1-hop unicast' routing
	route->SetGateway (ifaceAddr.GetBroadcast ());
	route->SetSource (ifaceAddr.GetLocal ());
	route->SetOutputDevice (dev);

	if (oif != 0 && route->GetOutputDevice () != oif)
	{
		NS_LOG_DEBUG ("Output device doesn't match. Dropped.");
		sockerr = Socket::ERROR_NOROUTETOHOST;
		return Ptr<Ipv4Route> ();
	}

	//std::cout << "[" << Simulator::Now ().GetSeconds () << "] " << ifaceAddr.GetLocal () << ": Route Output: returned broadcast route for packet " << p->GetUid () << std::endl;
	return route;
}

bool
RoutingProtocol::RouteInput (Ptr<const Packet> p, const Ipv4Header &header,
                             Ptr<const NetDevice> idev, UnicastForwardCallback ucb,
                             MulticastForwardCallback mcb, LocalDeliverCallback lcb, ErrorCallback ecb)
{
	NS_ASSERT (m_ipv4 != 0);
	NS_ASSERT (p != 0);
	// Check if input device supports IP
	NS_ASSERT (m_ipv4->GetInterfaceForDevice (idev) >= 0);
	int32_t iif = m_ipv4->GetInterfaceForDevice (idev);

	Ipv4InterfaceAddress ifaceAddr = m_ipv4->GetAddress (1, 0);		// assuming only one device, and only one address on it
	Ipv4Address dst = header.GetDestination ();
	Ipv4Address origin = header.GetSource ();

	// Duplicate of own packet
	if (origin == ifaceAddr.GetLocal ()) {
		if (idev == m_lo) {
			// If packet is sent from the loopback device
			NS_LOG_LOGIC ("Unicast local delivery to myself... (loopback)");
			lcb (p, header, iif);
			return true;
		} else {
			// Just ignore duplicates
			return true;
		}
	}

	// PFR is not a multicast routing protocol
	if (dst.IsMulticast ())
	{
		return false;
	}

	// Check for duplicates
    if (m_dpd.IsDuplicate (p, header))
      {
        NS_LOG_DEBUG ("Duplicated packet " << p->GetUid () << " from " << origin << ". Drop.");
        return true;
      }

    // Broadcast local delivery/forwarding (TODO, eventually, e.g. in a similar way as in AODV...)
    // Broadcast local delivery/forwarding
    if (dst == ifaceAddr.GetBroadcast () || dst.IsBroadcast ())
    {
    	// Make a copy of the packet, and deliver the original packet
    	Ptr<Packet> packet = p->Copy ();
    	if (lcb.IsNull () == false)
    	{
    		NS_LOG_LOGIC ("Broadcast local delivery to " << dst);
    		lcb (p, header, iif);
    		// Fall through to additional processing
    	}
    	else
    	{
    		NS_LOG_ERROR ("Unable to deliver packet locally due to null callback " << p->GetUid () << " from " << origin);
    		ecb (p, header, Socket::ERROR_NOROUTETOHOST);
    	}
    	// If broadcasting is not enabled, do not forward
    	if (!EnableBroadcast)
    	{
    		return true;
    	}
    	// Forward the broadcast
    	double tmp = UniformVariable().GetValue (0.0, 1.0);
    	if (tmp < BroadcastProbability) {
            if (header.GetTtl () > 1)
              {
                NS_LOG_LOGIC ("Forward broadcast. TTL " << (uint16_t) header.GetTtl ());

            	// Send via ucb callback using broadcasting
            	Ptr<Ipv4Route> route;
            	route = Create<Ipv4Route> ();
            	route->SetDestination (header.GetDestination ());
            	//route->SetGateway (header.GetDestination ());		// TODO temporary solution for '1-hop unicast' routing
            	route->SetGateway (ifaceAddr.GetBroadcast ());
            	route->SetSource (ifaceAddr.GetLocal ());
            	route->SetOutputDevice (m_ipv4->GetNetDevice (1));
            	//std::cout << "[" << Simulator::Now ().GetSeconds () << "] " << ifaceAddr.GetLocal () << ": Route Input, packet " << p->GetUid () << " forwarded. TTL " << (uint16_t) header.GetTtl () << std::endl;
    //        	ucb (route, p, header);

            	// Delayed callback
            	Simulator::Schedule (Time(0.005*MilliSeconds (UniformVariable ().GetInteger (0, 100))),&RoutingProtocol::DelayedUcbCallback, this, ucb,route,packet,header);
              }
            else
              {
                NS_LOG_DEBUG ("TTL exceeded. Drop packet " << packet->GetUid ());
              }
    	} else {
    		//std::cout << "[" << Simulator::Now ().GetSeconds () << "] " << ifaceAddr.GetLocal () << ": Route Input, packet " << p->GetUid () << " not forwarded." << std::endl;
    	}
    	return true;
    }



	// Unicast local delivery
	if (m_ipv4->IsDestinationAddress (dst, iif))
	{
//		if (dst == ifaceAddr.GetBroadcast())
//			std::cout << "[" << Simulator::Now ().GetSeconds () << "] broadcast detected" << std::endl;
		if (lcb.IsNull () == false)
		{
			NS_LOG_LOGIC ("Unicast local delivery to " << dst);
			lcb (p, header, iif);
		}
		else
		{
			NS_LOG_ERROR ("Unable to deliver packet locally due to null callback " << p->GetUid () << " from " << origin);
			ecb (p, header, Socket::ERROR_NOROUTETOHOST);
		}
		return true;
	}

	// Forwarding
	double tmp = UniformVariable().GetValue (0.0, 1.0);
	if (tmp < BroadcastProbability) {
        if (header.GetTtl () > 1)
          {
            NS_LOG_LOGIC ("Forward broadcast. TTL " << (uint16_t) header.GetTtl ());

        	// Send via ucb callback using broadcasting
        	Ptr<Ipv4Route> route;
        	route = Create<Ipv4Route> ();
        	route->SetDestination (header.GetDestination ());
        	//route->SetGateway (header.GetDestination ());		// TODO temporary solution for '1-hop unicast' routing
        	route->SetGateway (ifaceAddr.GetBroadcast ());
        	route->SetSource (ifaceAddr.GetLocal ());
        	route->SetOutputDevice (m_ipv4->GetNetDevice (1));
        	//std::cout << "[" << Simulator::Now ().GetSeconds () << "] " << ifaceAddr.GetLocal () << ": Route Input, packet " << p->GetUid () << " forwarded. TTL " << (uint16_t) header.GetTtl () << std::endl;
//        	ucb (route, p, header);

        	// Delayed callback
        	Simulator::Schedule (MicroSeconds (5*UniformVariable ().GetInteger (0, 100)),&RoutingProtocol::DelayedUcbCallback, this, ucb,route,p,header);
          }
        else
          {
            NS_LOG_DEBUG ("TTL exceeded. Drop packet " << p->GetUid ());
          }
	} else {
		//std::cout << "[" << Simulator::Now ().GetSeconds () << "] " << ifaceAddr.GetLocal () << ": Route Input, packet " << p->GetUid () << " not forwarded." << std::endl;
	}
	return true;
}

void
RoutingProtocol::DelayedUcbCallback(UnicastForwardCallback ucb,Ptr<Ipv4Route> route, Ptr<const Packet> p, const Ipv4Header &header) {
//	Ipv4InterfaceAddress ifaceAddr = m_ipv4->GetAddress (1, 0);		// assuming only one device, and only one address on it
//	std::cout << "[" << Simulator::Now ().GetSeconds () << "] " << ifaceAddr.GetLocal () << ": Route Input, packet " << p->GetUid () << " forwarded. TTL " << (uint16_t) header.GetTtl () << std::endl;
	ucb(route, p, header);
}

void
RoutingProtocol::NotifyInterfaceUp (uint32_t i)
{
//  NS_LOG_FUNCTION (this << m_ipv4->GetAddress (i, 0).GetLocal ());
//  Ptr<Ipv4L3Protocol> l3 = m_ipv4->GetObject<Ipv4L3Protocol> ();
//  if (l3->GetNAddresses (i) > 1)
//    {
//      NS_LOG_WARN ("AODV does not work with more then one address per each interface.");
//    }
//  Ipv4InterfaceAddress iface = l3->GetAddress (i, 0);
//  if (iface.GetLocal () == Ipv4Address ("127.0.0.1"))
//    return;
//
//  // Create a socket to listen only on this interface
//  Ptr<Socket> socket = Socket::CreateSocket (GetObject<Node> (),
//                                             UdpSocketFactory::GetTypeId ());
//  NS_ASSERT (socket != 0);
//  socket->SetRecvCallback (MakeCallback (&RoutingProtocol::RecvAodv, this));
//  socket->BindToNetDevice (l3->GetNetDevice (i));
//  socket->Bind (InetSocketAddress (Ipv4Address::GetAny (), AODV_PORT));
//  socket->SetAllowBroadcast (true);
//  socket->SetAttribute ("IpTtl", UintegerValue (1));
//  m_socketAddresses.insert (std::make_pair (socket, iface));
//
//  // Add local broadcast record to the routing table
//  Ptr<NetDevice> dev = m_ipv4->GetNetDevice (m_ipv4->GetInterfaceForAddress (iface.GetLocal ()));
//  RoutingTableEntry rt (/*device=*/ dev, /*dst=*/ iface.GetBroadcast (), /*know seqno=*/ true, /*seqno=*/ 0, /*iface=*/ iface,
//                                    /*hops=*/ 1, /*next hop=*/ iface.GetBroadcast (), /*lifetime=*/ Simulator::GetMaximumSimulationTime ());
//  m_routingTable.AddRoute (rt);
//
//  // Allow neighbor manager use this interface for layer 2 feedback if possible
//  Ptr<WifiNetDevice> wifi = dev->GetObject<WifiNetDevice> ();
//  if (wifi == 0)
//    return;
//  Ptr<WifiMac> mac = wifi->GetMac ();
//  if (mac == 0)
//    return;
//
//  mac->TraceConnectWithoutContext ("TxErrHeader", m_nb.GetTxErrorCallback ());
//  m_nb.AddArpCache (l3->GetInterface (i)->GetArpCache ());
	printf("NotifyInterfaceUp %i\n",i);
	return;
}

void
RoutingProtocol::NotifyInterfaceDown (uint32_t i)
{
//  NS_LOG_FUNCTION (this << m_ipv4->GetAddress (i, 0).GetLocal ());
//
//  // Disable layer 2 link state monitoring (if possible)
//  Ptr<Ipv4L3Protocol> l3 = m_ipv4->GetObject<Ipv4L3Protocol> ();
//  Ptr<NetDevice> dev = l3->GetNetDevice (i);
//  Ptr<WifiNetDevice> wifi = dev->GetObject<WifiNetDevice> ();
//  if (wifi != 0)
//    {
//      Ptr<WifiMac> mac = wifi->GetMac ()->GetObject<AdhocWifiMac> ();
//      if (mac != 0)
//        {
//          mac->TraceDisconnectWithoutContext ("TxErrHeader",
//                                              m_nb.GetTxErrorCallback ());
//          m_nb.DelArpCache (l3->GetInterface (i)->GetArpCache ());
//        }
//    }
//
//  // Close socket
//  Ptr<Socket> socket = FindSocketWithInterfaceAddress (m_ipv4->GetAddress (i, 0));
//  NS_ASSERT (socket);
//  socket->Close ();
//  m_socketAddresses.erase (socket);
//  if (m_socketAddresses.empty ())
//    {
//      NS_LOG_LOGIC ("No aodv interfaces");
//      m_htimer.Cancel ();
//      m_nb.Clear ();
//      m_routingTable.Clear ();
//      return;
//    }
//  m_routingTable.DeleteAllRoutesFromInterface (m_ipv4->GetAddress (i, 0));

	printf("NotifyInterfaceDown \n");
	return;
}

void
RoutingProtocol::NotifyAddAddress (uint32_t i, Ipv4InterfaceAddress address)
{
//  NS_LOG_FUNCTION (this << " interface " << i << " address " << address);
//  Ptr<Ipv4L3Protocol> l3 = m_ipv4->GetObject<Ipv4L3Protocol> ();
//  if (!l3->IsUp (i))
//    return;
//  if (l3->GetNAddresses (i) == 1)
//    {
//      Ipv4InterfaceAddress iface = l3->GetAddress (i, 0);
//      Ptr<Socket> socket = FindSocketWithInterfaceAddress (iface);
//      if (!socket)
//        {
//          if (iface.GetLocal () == Ipv4Address ("127.0.0.1"))
//            return;
//          // Create a socket to listen only on this interface
//          Ptr<Socket> socket = Socket::CreateSocket (GetObject<Node> (),
//                                                     UdpSocketFactory::GetTypeId ());
//          NS_ASSERT (socket != 0);
//          socket->SetRecvCallback (MakeCallback (&RoutingProtocol::RecvAodv,this));
//          socket->BindToNetDevice (l3->GetNetDevice (i));
//          // Bind to any IP address so that broadcasts can be received
//          socket->Bind (InetSocketAddress (Ipv4Address::GetAny (), AODV_PORT));
//          socket->SetAllowBroadcast (true);
//          m_socketAddresses.insert (std::make_pair (socket, iface));
//
//          // Add local broadcast record to the routing table
//          Ptr<NetDevice> dev = m_ipv4->GetNetDevice (
//              m_ipv4->GetInterfaceForAddress (iface.GetLocal ()));
//          RoutingTableEntry rt (/*device=*/ dev, /*dst=*/ iface.GetBroadcast (), /*know seqno=*/ true,
//                                            /*seqno=*/ 0, /*iface=*/ iface, /*hops=*/ 1,
//                                            /*next hop=*/ iface.GetBroadcast (), /*lifetime=*/ Simulator::GetMaximumSimulationTime ());
//          m_routingTable.AddRoute (rt);
//        }
//    }
//  else
//    {
//      NS_LOG_LOGIC ("AODV does not work with more then one address per each interface. Ignore added address");
//    }

	std::cout << "NotifyAddAddress: " << address << " (interface " << i << ")" <<  std::endl;
	return;
}

void
RoutingProtocol::NotifyRemoveAddress (uint32_t i, Ipv4InterfaceAddress address)
{
//  NS_LOG_FUNCTION (this);
//  Ptr<Socket> socket = FindSocketWithInterfaceAddress (address);
//  if (socket)
//    {
//      m_routingTable.DeleteAllRoutesFromInterface (address);
//      m_socketAddresses.erase (socket);
//      Ptr<Ipv4L3Protocol> l3 = m_ipv4->GetObject<Ipv4L3Protocol> ();
//      if (l3->GetNAddresses (i))
//        {
//          Ipv4InterfaceAddress iface = l3->GetAddress (i, 0);
//          // Create a socket to listen only on this interface
//          Ptr<Socket> socket = Socket::CreateSocket (GetObject<Node> (),
//                                                     UdpSocketFactory::GetTypeId ());
//          NS_ASSERT (socket != 0);
//          socket->SetRecvCallback (MakeCallback (&RoutingProtocol::RecvAodv, this));
//          // Bind to any IP address so that broadcasts can be received
//          socket->Bind (InetSocketAddress (Ipv4Address::GetAny (), AODV_PORT));
//          socket->SetAllowBroadcast (true);
//          m_socketAddresses.insert (std::make_pair (socket, iface));
//
//          // Add local broadcast record to the routing table
//          Ptr<NetDevice> dev = m_ipv4->GetNetDevice (m_ipv4->GetInterfaceForAddress (iface.GetLocal ()));
//          RoutingTableEntry rt (/*device=*/ dev, /*dst=*/ iface.GetBroadcast (), /*know seqno=*/ true, /*seqno=*/ 0, /*iface=*/ iface,
//                                            /*hops=*/ 1, /*next hop=*/ iface.GetBroadcast (), /*lifetime=*/ Simulator::GetMaximumSimulationTime ());
//          m_routingTable.AddRoute (rt);
//        }
//      if (m_socketAddresses.empty ())
//        {
//          NS_LOG_LOGIC ("No aodv interfaces");
//          m_htimer.Cancel ();
//          m_nb.Clear ();
//          m_routingTable.Clear ();
//          return;
//        }
//    }
//  else
//    {
//      NS_LOG_LOGIC ("Remove address not participating in AODV operation");
//    }

	printf("NotifyRemoveAddress \n");
	return;

}


}  // namespace pfr
}  // namespace ns3
